/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : TSP_GA.cc                                                */
/*                                                                       */
/*       Date : November  6, 1997                                        */
/* Updated at : November  6, 1997                                        */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include "global.h"
#include "gene.h"
#include "chromo.h"
#include "pop.h"
#include "TSP_GA.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <time.h>


/*
//
*/

type_TSP_GA* defineTSP_GA( char *_dataFile,int _popSize, int _generations, 
            float _crossover, enum cross_t _crossoverType, float  _mutation, 
            enum delete_t _deletionType, int _elitismValue, 
            enum fitness_t  _fitnessType, eval_t _fitnessParam, int _seed)
{
   type_TSP_GA* this = (type_TSP_GA* ) calloc(1,sizeof(type_TSP_GA));

   if(this!=NULL)
   {

         this->dataFile = (char*) calloc(MAX_LINE_LENGTH,sizeof(char));
         strcpy(this->dataFile,_dataFile);

          this->popSize = _popSize;
      this->generations = _generations;
        this->crossover = _crossover ;
    this->crossoverType = _crossoverType ;
         this->mutation = _mutation ; 
     this->deletionType = _deletionType ;
     this->elitismValue = _elitismValue ;
      this->fitnessType = _fitnessType ;
     this->fitnessParam = _fitnessParam ;
             this->seed = _seed ;
         this->genePool = NULL;
     this->chromoLength = 0;
            this->myPop = NULL;

        this->edgeGraph = NULL;

   }

   return this;

}// end of 

/*
//
*/

void replaceWorstChromosomeinTSPWith(type_TSP_GA* this, type_Chromosome* newChro) 
{
   if(this->myPop!=NULL) replaceWorstChromosomeWith(this->myPop,newChro);

}// end of replaceWorstChromosomeWith()

/*
//
*/

void setCrossoverProbability(type_TSP_GA* this, float val) 
{ this->crossover = val; }

/*
//
*/

void setMutationProbability(type_TSP_GA* this, float val) 
{ this->mutation = val; }

/*
//
*/

void setElitismValueu(type_TSP_GA* this,float val) 
{ this->elitismValue = val; }

/*
//
*/

void setRandomSeed(type_TSP_GA* this, int val) 
{

  if (val == 0)
    {
      time_t tv;
      time(&tv);
      this->seed = tv ;
    }
  else
    this->seed = val;

  srand(this->seed);

  printf("Using seed: %f \n",this->seed);

}// end of setRandomSeed()

/*
//
*/

void printPopulation(type_TSP_GA* this) 
{

   int i=0;
   type_Population* pop = this->myPop;
  
   if(pop!=NULL)
   {
      printf(" Population Size : %d \n",pop->popSize);
      printf(" Chromosome Length : %d \n ", pop->chromoLength);
     
      for(i=0;i<pop->popSize;i++)
      {
         printf(" %3d ",i);
         printChromosome(pop->myChromoPool[i]);
         printf("\n");
      }

   }

}// end of printPopulation

/*
//
*/

void printReport(type_TSP_GA* this) 
{
    printf(" TSP GA Parameters \n");
    printf("%20s %s\n"," Data File",this->dataFile);
    printf("%20s %d \n"," Population size",this->popSize);
    printf("%20s %d \n"," Max Generation",this->generations);
    printf("%20s %f \n"," Crossover prob ",this->crossover);
    printf("%20s "," Crossover type ");
    printf("%s\n",(this->crossoverType==0?"Fixed Position":"Fixed Position Sequential "));

    printf("%20s %f \n"," Mutation prob ",this->mutation);
    printf("%20s "," Deletion type ");
    printf("%s\n",(this->deletionType==deleteAll?"Generational":"Steady State"));

    printf("%20s %f \n"," Elitism value",this->elitismValue);
    printf("%20s "," Fitness type ");

    if(this->fitnessType==0) printf("%s\n","evaluation");
    else if(this->fitnessType==1) printf("%s\n","windowing");
    else if(this->fitnessType==2) printf("%s\n","linearNorm");
    else printf("%s\n","Unknown");

    printf("%20s %f \n"," Fitness param",this->fitnessParam);
    printf("%20s %f \n"," Seed",this->seed);
    printf("%20s %d \n"," Chromosome Length",this->chromoLength);

}// end of printReport()

/*
//
*/

type_Chromosome* getBestResult(type_TSP_GA* this) 
{
   return(bestChromo(this->myPop));

}// end of getBestResult()

/*
//
*/

void printBestResult(type_TSP_GA* this) 
{
   printf(" Best Chromosome : ");
   printChromosome(getBestResult(this));
   printf("\n");

}// end of printBestResult()

/*
//
*/

void printGenePool(type_TSP_GA* this)
{
  int i;

  for (i = 0; i < this->chromoLength; i++) 
  { 
     printf(" %3d ",i);
     printGene(*this->genePool[i]);
     printf("\n");
  }

}// end of printGenePool()

/*
//
*/

void runTimes(type_TSP_GA* this, unsigned iterationNumber)
{

   type_Population *myNewPop;
   int k;

   /*********************************************/
   /** If the there is no previous population,  */
   /** then define it now.                      */ 
   /*********************************************/
   printf("before define pop\n");
   if(this->myPop==NULL) 
      this->myPop = definePopulation(this->popSize, this->chromoLength, this->genePool);

   printf(" Population is created! \n");
   printPopulation(this);


   /******************************/
   /** Run the GA iterationTimes */
   /******************************/

   for (k = 0; k < iterationNumber; k++)
   {
       
      fitness(this->myPop, this->fitnessType, this->fitnessParam);

//      cout << "[" << k <<"] average: "
//	<< myPop->fitnessAverage()
//	  << "    best: " << myPop->bestFitness() << endl;

      myNewPop = newPop(this->myPop, this->fitnessType, this->fitnessParam, this->crossover,
			       this->crossoverType, this->mutation, this->deletionType,
			       this->elitismValue);

      if (this->deletionType == deleteAll)
      {
	  free(this->myPop);
	  this->myPop = myNewPop; 

      }// end of if

      printf(" %d th generation \n ",k);
      printPopulation(this);

   }// end of for

}// end of runTimes()

/*
*/

/*
// ret = 0 : OK
// ret = 1 : File can not be opened!
//
*/

readGenes(type_TSP_GA* this)
{

  char buf1[nameLength + 1];
  char buf2[nameLength + 1];
  char buf3[nameLength + 1];
  char buf[MAX_LINE_LENGTH];
  coord_t x,y;
  unsigned int i, j ;

  int ret = 0;

  FILE *fp = fopen(this->dataFile,"r");

  if (fp!=NULL)
  {

     /****************************************/
     /** Read number of elements on the file */
     /****************************************/
     fscanf(fp,"%d\n",&this->chromoLength);

     this->edgeGraph = defineEdgeGraph(this->chromoLength);
     this->genePool = (type_Gene**) calloc(this->chromoLength,sizeof(type_Gene*));

     /***********************/
     /** Read the data file */
     /***********************/

     for (i = 0; !feof(fp) && i < this->chromoLength ; i++)
     {

        fscanf(fp,"%s %f %f \n",buf1,&x,&y);

        this->genePool[i] = defineGene(buf1, x, y, i);

         addEdge(this->edgeGraph,*this->genePool[i]);

     }// end of for
     ret = 0;
     close(fp);
  }
  else
  {
      printf("Error: Couldn't find gene data file `%s'\n", this->dataFile);
      ret = 1;
  }

  if(ret) { free(this->genePool); 
	    this->genePool=NULL; }
  else
    for (i = 0; i < this->chromoLength; i++)
      for (j = 0; j < this->chromoLength; j++)
           calculateDistanceTo(this->genePool[i],this->genePool[j]);

  return ret;

}// end of readGenes

/*
//
*/

void printAllDistancesInGenePool(type_TSP_GA* this)
{
  int i, j;

  for (i = 0; i < this->chromoLength; i++)
    for (j = 0; j < this->chromoLength; j++)
      printf(" [ %3d,%3d ] %s - %s : %f \n", i,j,
              getName(*this->genePool[i]), getName(*this->genePool[j]), 
	      distanceTo(this->genePool[i],this->genePool[j]));

}// end of printAllDistancesInGenePool()

void disposeTSP_GA(type_TSP_GA* this)
{
   free(this->dataFile);
   free(this);
}
